#!/usr/bin/perl -w
#
#    Copyright 2001,2004 Greenwood Genetic Center
#
#    This file is part of Linkage Scripts.
#
#    Linkage Scripts is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    Linkage Scripts is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Linkage Scripts; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use strict;

use Linkage;

parseDatafile();
parsePedfile();

my(@outfiles) = ("lsp.log",
                 "lsp.stm",
                 "lsp.tmp",
                 "final.dat",
                 "stream.dat",
                 "outfile.dat",
                 "recfile.dat",
                 "final.out",
                 "stream.out");

my(@tmpfiles) = ("datafile.dat",
                 "ipedfile.dat",
                 "pedfile.dat",
                 "speedfile.dat",
                 "tempdat.dat",
                 "tempped.dat");

my($disease_locus) = 1;
my($runnumber) = 1;
my(@families) = ();

my($m1, $m2, $m3, $m4, $m5, $firstlocus, $firstlocus_stop, $newrun, $sendto,
   $locusorder, $printname, $skipprint, $currentmarker, $pedentry, $newfamily,
   $title, $tempfile);

# Determine the title from the families in the pedfile
foreach $pedentry (@pedinfo) {
  $newfamily = 1; #true
  foreach(@families) {
    if($pedentry->[@$pedentry - 2] eq $_) {
      $newfamily = 0; #false
    }
  }
  if($newfamily == 1) {
    push(@families, $$pedentry[@$pedentry - 2]);
  }
}

if(@families == 1) {
  $title = "Family: $families[0]";
} else {
  $title = "Families: " . join(", ", @families,);
}

# see if we should run in debug mode (e.g. print the output of utilities to
# the screen)
if(exists($ARGV[0]) && $ARGV[0] eq "DEBUG") {
  $sendto = "";
} else {
  $tempfile = "/tmp/linkage" . int(rand(10 ** 10));
  $sendto = " &> $tempfile";
}

# make sure we have enough markers to run this script
if($datainfo{"nlocus"} < 5) {
  print "At least five markers are required to run this script.\n";
  exit(1);
}

# delete the known temporary files, the previous output files, and the
# temporary wildcard files
foreach((@outfiles,
         @tmpfiles,
         <*.tpd>,
         <*.tdt>,
         <*.mpd>,
         <*.mdt>,
	 <checkpoint*>,
	 <script*>,
	 <outf*>,
	 <main*>)) {

  if(-f $_) {
    unlink($_)
        or die "Couldn't delete $_: $!\n";
  }
}

# calculate the value that the script should stop at
# total no. of loci - 3 (since there are 4 markers in each test)
$firstlocus_stop = $datainfo{"nlocus"} - 3;

for($firstlocus = 2; $firstlocus <= $firstlocus_stop; $firstlocus++) {
  # $firstlocus refers to the chromosome order, so look this up
  $m1 = $datainfo{"order"}->[$firstlocus - 1];
  $m2 = $datainfo{"order"}->[$firstlocus];
  $m3 = $datainfo{"order"}->[$firstlocus + 1];
  $m4 = $datainfo{"order"}->[$firstlocus + 2];

  $newrun = 0;
  #this is a fresh set of markers; let lsp regenerate the pedfile and datafile

  # run linkmap between the first and second locus to calculate baseline
  # likelihoods
  runLinkmap($disease_locus, $m1, $m2, $m3, $m4,
      0.5, @{$datainfo{"recomb"}}[$firstlocus - 1],
      @{$datainfo{"recomb"}}[$firstlocus],
      @{$datainfo{"recomb"}}[$firstlocus + 1],
      0);

  if($firstlocus == 2) { # if this is the first locus set
    runLinkmap($m1, $disease_locus, $m2, $m3, $m4,
        0, @{$datainfo{"recomb"}}[$firstlocus - 1],
        @{$datainfo{"recomb"}}[$firstlocus],
        @{$datainfo{"recomb"}}[$firstlocus + 1],
        @{$datainfo{"recomb"}}[$firstlocus - 1]);
  }

  runLinkmap($m1, $m2, $disease_locus, $m3, $m4,
      @{$datainfo{"recomb"}}[$firstlocus - 1], 0,
      @{$datainfo{"recomb"}}[$firstlocus],
      @{$datainfo{"recomb"}}[$firstlocus + 1],
      @{$datainfo{"recomb"}}[$firstlocus]);

  if($firstlocus == $firstlocus_stop) { # if this is the last locus set
    runLinkmap($m1, $m2, $m3, $disease_locus, $m4,
        @{$datainfo{"recomb"}}[$firstlocus - 1],
        @{$datainfo{"recomb"}}[$firstlocus], 0,
        @{$datainfo{"recomb"}}[$firstlocus + 1],
        @{$datainfo{"recomb"}}[$firstlocus + 1]);
    runLinkmap($m1, $m2, $m3, $m4, $disease_locus,
        @{$datainfo{"recomb"}}[$firstlocus - 1],
        @{$datainfo{"recomb"}}[$firstlocus],
        @{$datainfo{"recomb"}}[$firstlocus + 1], 0,
        0.5);
  }
}

# clean up our tempfiles and the temporary wildcard files
foreach((@tmpfiles,
         <*.tpd>,
         <*.tdt>,
         <*.mpd>,
         <*.mdt>,
	 <checkpoint*>,
	 <script*>,
	 <outf*>,
	 <main*>)) {
  if(-f $_) {
    unlink($_)
        or die "Couldn't delete $_: $!\n";
  }
}

# run linklods to calculate lod scores
if(system("echo | linklods" . $sendto) != 0) {
  dump_prog_out();
  die "\nlinkmap.pl:  linklods failed to run properly; see message above\n";
}


# Print the results
open(LODSCORE, "<final.lod")
    or die "Couldn't open final.lod: $!\n";

open(RESULTS, ">linkmap.csv")
    or die "Couldn't open linkmap.csv: $!\n";

$printname = 0;
$skipprint = 0;
$currentmarker = 1;

print RESULTS "$title\n\n";

print RESULTS "Markers,Distances,,,,LOD Score\n";

foreach(<LODSCORE>) {
  if($skipprint == 0) {
    if($_ =~ m/THETAS line with 0\.5 found\./) {
      $skipprint = 1; # don't print these scores because they are only used to
    }                 # calculate the baseline likelihoods.

    if($_ =~ m/THETAS\s+([\d.]+)\s+([\d.]+)\s+([\d.]+)\s+([\d.]+)/) {
      if($printname == 0) {
        print RESULTS "$locusorder,$1,$2,$3,$4,"; #<- print the locus order and
      } else {                                    # keep the value of $printname
        print RESULTS ",$1,$2,$3,$4,";            # as zero so the actual locus
      }                                           # name will be printed later
    }
    
    if($_ =~ m/TOTALS\s+([\d.-]+)/) {
      if($printname == 0) {
        print RESULTS "$1,," .
            $datainfo{"names"}->[$datainfo{"order"}->[$currentmarker++] - 1] .
            "\n";
        $printname = 1; #don't print the names anymore
      } else {
        print RESULTS "$1\n";
      }
    }
  }
  
  if($_ =~ m/Locus Order:  (\d+ \d+ \d+ \d+ \d+)/) {
    $locusorder = $1; # save the current locus order
    $skipprint = 0; # don't skip the next print
    $printname = 0; # print the name of this locus later
  }
}

print RESULTS "\n\nProcessed on " . localtime(time()) . "\n";

close(RESULTS)
    or warn "Couldn't close linkmap.csv: $!\n";

close(LODSCORE)
    or warn "Couldn't close final.lod: $!\n";

sub runLinkmap {
  printf("Run %2d- LINKMAP: %2d(%-7s)--%2d(%-7s)--%2d(%-7s)--%2d(%-7s)--" .
      "%2d(%-7s)\n",
      $runnumber++, $_[0], $datainfo{"names"}->[$_[0] - 1], $_[1],
      $datainfo{"names"}->[$_[1] - 1], $_[2], $datainfo{"names"}->[$_[2] - 1],
      $_[3], $datainfo{"names"}->[$_[3] - 1], $_[4],
      $datainfo{"names"}->[$_[4] - 1]);

  printf("                          %2.5f      %2.5f      %2.5f      %2.5f\n\n",
         $_[5], $_[6], $_[7], $_[8]);

  copy("pedin.dat", "pedin.tpd")
       or die "Couldn't copy pedin.dat to pedin.tpd: $!\n";
  copy("datain.dat", "datain.tdt")
       or die "Couldn't copy datain.dat to datain.tdt: $!\n";

  if($newrun == 0) {
    $newrun = 1; # we've run lsp once on this marker set
    if(system("lsp " .
              "linkmap " .
              "pedin.tpd datain.tdt " .
              "5 " .
              "$_[0]  $_[1]  $_[2]  $_[3]  $_[4] " .
              "1 $_[9] 5 0 " .
              "$_[5] $_[6] $_[7] $_[8] " .
              $sendto) != 0) {

      dump_prog_out();
      die "\nlinkmap.pl:  lsp failed to run properly; see output above\n";
    }
  } else {
    if(system("lsp " .
              "linkmap " .
              "= datain.tdt " .
              "5 " .
              "$_[0]  $_[1]  $_[2]  $_[3]  $_[4] " .
              "1 $_[9] 5 0 " .
              "$_[5] $_[6] $_[7] $_[8] " .
              $sendto) != 0) {

      dump_prog_out();
      die "\nlinkmap.pl:  lsp failed to run properly; see output above\n";
    }
  }

  system("cat lsp.log >> final.out");
  system("cat lsp.stm >> stream.out");

  if(system("unknown" . $sendto) != 0) {
    dump_prog_out();
    die "\nlinkmap.pl:  unknown failed to run properly; see output above\n";
  }

  if(system("linkmap" . $sendto) != 0) {
    dump_prog_out();
    die "\nlinkmap.pl:  linkmap failed to run properly; see output above\n";
  }

  system("cat outfile.dat >> final.out");
  system("cat stream.dat >> stream.out");
}

print "Analysis Done!\n";


##########################################################################
##  copy($from, $to)
##    copies a $from to $to
##########################################################################
sub copy {
  open(FINPUT, "<$_[0]") or die "Couldn't open $_[0]: $!\n";
  open(FOUTPUT, ">$_[1]") or die "Couldn't open $_[1]: $!\n";
  foreach(<FINPUT>) {
    print FOUTPUT $_;
  }
  close(FINPUT) or warn "Couldn't close $_[0]: $!\n";
  close(FOUTPUT) or warn "Couldn't close $_[1]: $!\n";
}


##########################################################################
##  dump_prog_out()
##    dumps the error message of the last program run if captured to a file
##########################################################################
sub dump_prog_out {
    if($tempfile)  {
	open(LINKOUT, "$tempfile") or die "Couldn't open $tempfile: $!\n";

	foreach(<LINKOUT>) {
	    print $_;
	}

	close(LINKOUT) or warn "Unable to close $tempfile: $!\n";
    }
}

END {
    if($tempfile) {
	unlink($tempfile) or warn "unable to unlink $tempfile: $!\n";
    }
}
